/* -*- mode: c++; tab-width: 4; c-basic-offset: 4 -*- */
/*
 * Copyright (C) 2005 Atmark <atmarkat _AT_ msn _DOT_ com>
 *                    AGAWA Koji <i _AT_ atty _DOT_ jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef ACTION_H_INCLUDED
#define ACTION_H_INCLUDED

#include <qstring.h>
#include <qcstring.h>
#include <qlist.h>
#include <qmap.h>
#include "configuration.h"

/*! アクションの抽象を提供します。

 このクラスは直接作成せずに ActionManager クラスを通して利用してください。

 \sa Quasar::keyPressEvent(QKeyPressEvent *)
 */
class Action
{
public:
	Action() : accel_(-1) {}
	Action(const QCString& name, int accel);

	//! アクションの名前を返します。
	QCString name() const { return name_; }
	//! アクションの翻訳された名前を返します。
	QString translatedName() const;
	//! アクションのアクセラレータキーを返します。
	int accel() const { return accel_; }

	int operator<(const Action& rhs) { return name_ < rhs.name(); }
	int operator==(const Action& rhs) { return name_ == rhs.name(); }

private:
	QCString name_;
	int accel_;
};

/*! アクションを管理するクラスです。

 ZPlayerの各アクション(再生、停止など)へのユーザーからのアクセスはこのクラスで
抽象化されます。

* アクションを新しく追加するには

 1. ActionManager::actionDecls_ 配列にアクションの定義を追加します。

  ホットキー設定ダイアログのアクション一覧は、この配列に定義された順番でリスト
 されます。それを考慮して定義を追加する位置を決めてください。
  そのとき、IDが一意になるように注意してください。

 2. ActionManager::createDefaultHotKeyMap() にアクションのデフォルトホットキー
    の定義を追加します。

 3-a. アクションにするウィジェットのアクセラレータキーを、Action::accel()の値に
      設定します。

<example>
	QPushButton *button = new QPushButton("Pause", this);
	this->setAccel(g_ActionManager.actionByName("Playback/Pause").accel());
</example>

 3-b. アクセラレータキーの設定を持たないもの(関数など)をアクションにする場合は、
      QActionを使ってください。

<example>
	QAction *a = new QAction(this, "Playback/Pause");
	connect(a, SIGNAL(activated()), player, SLOT(pause()));
	a->setAccel(g_ActionManager.actionByName("Playback/Pause").accel());
</example>
 */
class ActionManager
{
public:
	static ActionManager& singleton();

	//! アクション名 \a name に対応するアクションオブジェクトを返します。
	const Action& actionByName(const QCString& name) const;
	//! アクセラレータキー \a accel に対応するアクションオブジェクトを返します。
	const Action& actionByAccel(int accel) const;

	//! 全てのアクションをのアクションオブジェクトのリストで返します。
	QList<Action> allActions() const;

	//! デフォルトのホットキー定義を返します。
	Configuration::HotKeyMap defaultHotKey(Configuration::HotKeyContext context) const;

private:
	ActionManager();

	//! デフォルトのホットキー定義です。
	Configuration::HotKeyMap defaultHotKeyMap_[Configuration::HOTKEY_CONTEXT_MAX];

	//! デフォルトのホットキー定義を作成します。
	void createDefaultHotKeyMap();

	//! createDefaultHotKeyMap() から使われる、便宜のためのメンバ関数です。
	void addHotKey(const QCString &name, int key, int c1,
				   int c2 = Configuration::HOTKEY_CONTEXT_MAX,
				   int c3 = Configuration::HOTKEY_CONTEXT_MAX,
				   int c4 = Configuration::HOTKEY_CONTEXT_MAX);

	struct ActionDecl {
		int id;
		const char *name;
	};
	static const ActionDecl actionDecls_[];

	typedef QMap<QCString, Action> ActionMap;
	ActionMap actionMap_;

	Action nullAction_;
};
#define g_ActionManager ActionManager::singleton()

#endif // ACTION_H_INCLUDED
